from sqlalchemy import Column, DateTime, String
from sqlalchemy.orm import declarative_mixin

from ulid import ULID
import datetime


def generate_primary_key() -> str:
    return str(ULID())


@declarative_mixin
class DBEntityMixin():
    id = Column(
        String(36), primary_key=True, index=True, default=generate_primary_key
    )
    created_at = Column(DateTime, default=datetime.datetime.utcnow)
    created_by = Column(String(36), nullable=True)
    updated_at = Column(DateTime, nullable=True)
    updated_by = Column(String(36), nullable=True)



from zrb import (
    runner, CmdTask, BoolInput, StrInput, HTTPChecker
)

build = CmdTask(
    name='build',
    description='Build Zrb',
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Build zrb distribution"',
        'rm -Rf ${ZRB_PROJECT_DIR}/dist',
        'git add . -A',
        'flit build',
    ],
)
runner.register(build)

publish = CmdTask(
    name='publish',
    description='Publish zrb to pypi',
    upstreams=[build],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Publish zrb to pypi"',
        'flit publish --repository pypi',
    ]
)
runner.register(publish)

publish_test = CmdTask(
    name='publish-test',
    description='Publish zrb to testpypi',
    upstreams=[build],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Publish zrb to testpypi"',
        'flit publish --repository testpypi',
    ]
)
runner.register(publish_test)

install_symlink = CmdTask(
    name='install-symlink',
    description='Install Zrb as symlink',
    upstreams=[build],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Install zrb"',
        'flit install --symlink',
    ]
)
runner.register(install_symlink)

test = CmdTask(
    name='test',
    description='Run zrb test',
    inputs=[
        StrInput(
            name='test',
            shortcut='t',
            description='Specific test case (i.e., test/file.py::test_name)',
            prompt='Test (i.e., test/file.py::test_name)',
            default=''
        ),
        BoolInput(
            name='parallel',
            description='Whether doing parallel testing or not',
            prompt='Parallel testing?',
            default=True
        )
    ],
    upstreams=[install_symlink],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Perform test"',
        'pytest {{ "-n auto " if input.parallel else "" }}--ignore-glob=**/template/**/test --ignore=playground --cov=zrb --cov-report=html --cov-report=term --cov-report=term-missing {{input.test}}'  # noqa
    ],
    retry=0,
    checking_interval=1
)
runner.register(test)

serve_test = CmdTask(
    name='serve-test',
    description='Serve zrb test result',
    inputs=[
        StrInput(
            name='port',
            shortcut='p',
            description='Port to serve coverage result',
            prompt='Serve coverage on port',
            default='9000'
        )
    ],
    upstreams=[test],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Serve coverage report"',
        'python -m http.server {{input.port}} --directory ${ZRB_PROJECT_DIR}/htmlcov',  # noqa
    ],
    checkers=[
        HTTPChecker(port='{{input.port}}')
    ],
    retry=0,
    checking_interval=0.3
)
runner.register(serve_test)

playground = CmdTask(
    name='playground',
    upstreams=[install_symlink],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Remove playground"',
        'sudo rm -Rf playground',
        'echo "🤖 Create playground"',
        'cp -R playground-template playground',
        'cd ${ZRB_PROJECT_DIR}/playground',
        'echo "🤖 Generate project"',
        './generate-project.sh',
        'echo "🤖 Change to playground directory:"',
        'echo "      cd ${ZRB_PROJECT_DIR}/playground"',
        'echo "🤖 Or playground project directory:"',
        'echo "      cd ${ZRB_PROJECT_DIR}/playground/my-project"',
        'echo "🤖 And start hacking around. Good luck :)"',
    ],
    retry=0,
    preexec_fn=None
)
runner.register(playground)



from zrb.helper.codemod.add_function_call import add_function_call


def test_import_module():
    code = add_function_call(
        code='\n'.join([
            'def add(a, b):',
            '    return a + b'
        ]),
        function_name='add',
        parameters=['4', '5']
    )
    assert code == '\n'.join([
        'def add(a, b):',
        '    return a + b',
        'add(4, 5)'
    ])



from module.log.core.historical_repo_model import HistoricalRepoModel
from module.snake_module_name.schema.snake_entity_name import (
    PascalEntityName, PascalEntityNameData, PascalEntityNameResult
)


class PascalEntityNameModel(
    HistoricalRepoModel[PascalEntityName, PascalEntityNameData, PascalEntityNameResult]
):
    schema_result_cls = PascalEntityNameResult
    log_entity_name = 'snake_entity_name'



from typing import Any, List, Mapping, Optional, TypeVar, Type
from pydantic import BaseModel
from sqlalchemy import or_
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql._typing import _ColumnExpressionArgument
from core.repo.search_filter import SearchFilter
from core.repo.repo import Repo

from ulid import ULID
import datetime
import logging

Schema = TypeVar('Schema', bound=BaseModel)
SchemaData = TypeVar('SchemaData', bound=BaseModel)
DBEntity = TypeVar('DBEntity', bound=Any)


class DBRepo(Repo[Schema, SchemaData]):
    schema_cls: Type[Schema]
    db_entity_cls: Type[DBEntity]

    def __init__(
        self, logger: logging.Logger, engine: Engine,
    ):
        self.logger = logger
        self.engine = engine
        self.db_entity_attribute_names: List[str] = dir(self.db_entity_cls)
        self._keyword_fields: Optional[List[InstrumentedAttribute]] = None

    async def get_by_id(self, id: str) -> Schema:
        '''
        Find a record by id.
        '''
        db = self._create_db_session()
        try:
            search_filter = self.db_entity_cls.id == id
            db_entity = self._get_one_by_criterion(db, search_filter)
            return self._db_entity_to_schema(db, db_entity)
        finally:
            db.close()

    async def get(
        self, search_filter: Optional[SearchFilter] = None,
        limit: int = 100, offset: int = 0
    ) -> List[Schema]:
        '''
        Find multiple records by keyword with limit and offset.
        '''
        db = self._create_db_session()
        try:
            search_filter = self._ensure_search_filter(search_filter)
            criterion = self._search_filter_to_criterion(search_filter)
            db_entities = self._get_by_criterion(db, criterion, limit, offset)
            return [
                self._db_entity_to_schema(db, db_entity)
                for db_entity in db_entities
            ]
        finally:
            db.close()

    async def count(self, search_filter: Optional[SearchFilter] = None) -> int:
        '''
        Count records by keyword.
        '''
        db = self._create_db_session()
        try:
            search_filter = self._ensure_search_filter(search_filter)
            criterion = self._search_filter_to_criterion(search_filter)
            return self._count_by_criterion(db, criterion)
        finally:
            db.close()

    async def insert(self, data: SchemaData) -> Schema:
        '''
        Insert a new record.
        '''
        db = self._create_db_session()
        try:
            db_entity = self.db_entity_cls(
                **self._schema_data_to_db_entity_map(db, data),
            )
            if 'id' in self.db_entity_attribute_names:
                new_id = self.generate_id()
                db_entity.id = new_id
            if 'created_at' in self.db_entity_attribute_names:
                db_entity.created_at = datetime.datetime.utcnow()
            if 'updated_at' in self.db_entity_attribute_names:
                db_entity.updated_at = datetime.datetime.utcnow()
            db.add(db_entity)
            db.commit()
            db.refresh(db_entity)
            return self._db_entity_to_schema(db, db_entity)
        except Exception:
            self.logger.error(
                f'Error while inserting into {self.db_entity_cls} ' +
                f'with schema_data: {data}'
            )
            raise
        finally:
            db.close()

    def generate_id(self) -> str:
        return str(ULID())

    async def update(self, id: str, data: SchemaData) -> Schema:
        '''
        Update a record.
        '''
        db = self._create_db_session()
        try:
            db_entity = self._get_one_by_criterion(
                db, self.db_entity_cls.id == id
            )
            db_entity_map = self._schema_data_to_db_entity_map(db, data)
            for field, value in db_entity_map.items():
                if field == 'created_at' or field == 'created_by':
                    continue
                setattr(db_entity, field, value)
            if 'updated_at' in self.db_entity_attribute_names:
                db_entity.updated_at = datetime.datetime.utcnow()
            db.add(db_entity)
            db.commit()
            db.refresh(db_entity)
            return self._db_entity_to_schema(db, db_entity)
        except Exception:
            self.logger.error(
                f'Error while updating {self.db_entity_cls} ' +
                f'with id: {id}, schema_data: {data}'
            )
            raise
        finally:
            db.close()

    async def delete(self, id: str) -> Schema:
        '''
        Delete a record.
        '''
        db = self._create_db_session()
        try:
            db_entity = self._get_one_by_criterion(
                db, self.db_entity_cls.id == id
            )
            db.delete(db_entity)
            db.commit()
            return self._db_entity_to_schema(db, db_entity)
        except Exception:
            self.logger.error(
                f'Error while deleting {self.db_entity_cls} with id: {id}'
            )
            raise
        finally:
            db.close()

    def _create_db_session(self) -> Session:
        '''
        Return a db session.
        '''
        return Session(self.engine, expire_on_commit=False)

    def _get_by_criterion(
        self,
        db: Session, criterion: _ColumnExpressionArgument[bool],
        limit: int = 100, offset: int = 0
    ) -> List[DBEntity]:
        try:
            db_query = db.query(self.db_entity_cls).filter(
                criterion
            )
            if 'created_at' in self.db_entity_attribute_names:
                db_query = db_query.order_by(
                    self.db_entity_cls.created_at.desc()
                )
            return db_query.offset(offset).limit(limit).all()
        except Exception:
            self.logger.error(
                f'Error while getting {self.db_entity_cls} ' +
                f'with criterion: {criterion}, ' +
                f'limit: {limit}, offset: {offset}'
            )
            raise

    def _count_by_criterion(
        self, db: Session, criterion: _ColumnExpressionArgument[bool]
    ) -> int:
        try:
            return db.query(self.db_entity_cls).filter(
                criterion
            ).count()
        except Exception:
            self.logger.error(
                f'Error while counting for {self.db_entity_cls} ' +
                f'with criterion: {criterion}'
            )
            raise

    def _get_one_by_criterion(
        self, db: Session, criterion: _ColumnExpressionArgument[bool]
    ) -> DBEntity:
        try:
            db_entity = db.query(self.db_entity_cls).filter(
                criterion
            ).first()
            if db_entity is None:
                raise ValueError(
                    f'Not found: Cannot find a {self.db_entity_cls} ' +
                    f'with criterion: {criterion}'
                )
            return db_entity
        except Exception:
            self.logger.error(
                f'Error while getting a {self.db_entity_cls} ' +
                f'with criterion: {criterion}'
            )
            raise

    def _schema_data_to_db_entity_map(
        self, db: Session, schema_data: SchemaData
    ) -> Mapping[str, Any]:
        '''
        Convert entity_data into dictionary
        The result of this convertion is used for inserting/updating db_entity.
        '''
        entity_dict = schema_data.dict()
        return {
            field: value
            for field, value in entity_dict.items()
            if field in self.db_entity_attribute_names
        }

    def _db_entity_to_schema(self, db: Session, db_entity: DBEntity) -> Schema:
        '''
        Convert db_entity into schema.
        '''
        return self.schema_cls.from_orm(db_entity)

    def _search_filter_to_criterion(
        self, search_filter: SearchFilter
    ) -> _ColumnExpressionArgument[bool]:
        '''
        Return keyword filtering.
        The result is usually used to invoke find/count.
        '''
        keyword = search_filter.keyword
        like_keyword = '%{}%'.format(keyword) if keyword != '' else '%'
        keyword_criterion = [
            keyword_field.like(like_keyword)
            for keyword_field in self._get_keyword_fields()
        ]
        return or_(*keyword_criterion)

    def _get_keyword_fields(self) -> List[InstrumentedAttribute]:
        '''
        Return list of fields for keyword filtering
        '''
        if self._keyword_fields is not None:
            return self._keyword_fields
        self._keyword_fields = []
        for field_name in self.db_entity_attribute_names:
            field = getattr(self.db_entity_cls, field_name, None)
            if type(field) != InstrumentedAttribute:
                continue
            field_type = getattr(field, 'type', None)
            if field_type is None:
                continue
            str_field_type = str(field_type)
            if not (
                str_field_type.upper().startswith('VARCHAR') or
                str_field_type.upper().startswith('TEXT')
            ):
                continue
            self._keyword_fields.append(field)
        return self._keyword_fields

    def _ensure_search_filter(self, search_filter: Optional[SearchFilter]):
        if search_filter is None:
            return SearchFilter(keyword='', criterion={})
        return search_filter



from ..log import logger
import os
import re
import importlib.util
from ..accessories.color import colored

pattern = re.compile('[^a-zA-Z0-9]')


def load_module(script_path: str):
    if os.path.isfile(script_path):
        logger.info(colored(f'Load {script_path}', attrs=['dark']))
        module_name = pattern.sub('', script_path)
        spec = importlib.util.spec_from_file_location(module_name, script_path)
        # Create a module from the spec
        module = importlib.util.module_from_spec(spec)
        # Execute the module code
        spec.loader.exec_module(module)



from ...task.installer.factory import create_installer
from ...task_input.str_input import StrInput
from ...task_input.bool_input import BoolInput
from ...runner import runner
import os

dir_path = os.path.dirname(__file__)


# GVM
gvm_install_task = create_installer(
    name='gvm',
    template_dir=os.path.join(dir_path, 'gvm'),
    description='GVM provides interface to manage go version',
    inputs=[
        StrInput(
            name='go-default-version',
            description='Go default version',
            default='go1.14'
        )
    ],
    config_locations=['~/.bashrc', '~/.zshrc']
)
runner.register(gvm_install_task)

# PyEnv
pyenv_install_task = create_installer(
    name='pyenv',
    template_dir=os.path.join(dir_path, 'pyenv'),
    description='Simple Python version management',
    inputs=[
        StrInput(
            name='python-default-version',
            description='Python default version',
            default='3.9.0'
        )
    ],
    config_locations=['~/.bashrc', '~/.zshrc']
)
runner.register(pyenv_install_task)

# NVM
nvm_install_task = create_installer(
    name='nvm',
    template_dir=os.path.join(dir_path, 'nvm'),
    description=' '.join([
        'NVM allows you to quickly install and use different versions',
        'of node via the command line'
    ]),
    inputs=[
        StrInput(
            name='node-default-version',
            description='Node default version',
            default='node'
        )
    ],
    config_locations=['~/.bashrc', '~/.zshrc']
)
runner.register(nvm_install_task)

# sdkman
sdkman_install_task = create_installer(
    name='sdkman',
    template_dir=os.path.join(dir_path, 'sdkman'),
    description=' '.join([
        'SDKMAN! is a tool for managing parallel versions of multiple',
        'Software Development Kits on most Unix based systems'
    ]),
    inputs=[
        BoolInput(
            name='install-java',
            description='Install Java',
            prompt='Do you want to install Java?',
            default=True
        ),
        BoolInput(
            name='install-scala',
            description='Install Scala',
            prompt='Do you want to install Scala?',
            default=True
        )
    ],
    config_locations=['~/.bashrc', '~/.zshrc']
)
runner.register(sdkman_install_task)

# pulumi
pulumi_install_task = create_installer(
    name='pulumi',
    template_dir=os.path.join(dir_path, 'pulumi'),
    description='Universal infrastructure as code',
    config_locations=['~/.bashrc', '~/.zshrc']
)
runner.register(pulumi_install_task)

# AWS
aws_install_task = create_installer(
    name='aws',
    template_dir=os.path.join(dir_path, 'aws'),
    description='AWS CLI',
    ask_config_location=False
)
runner.register(aws_install_task)

# GCloud
gcloud_install_task = create_installer(
    name='gcloud',
    template_dir=os.path.join(dir_path, 'gcloud'),
    description='Gcloud CLI',
    ask_config_location=False
)
runner.register(gcloud_install_task)

# Tmux
tmux_install_task = create_installer(
    name='tmux',
    template_dir=os.path.join(dir_path, 'tmux'),
    description='Tmux',
    default_config_location='~/.tmux.conf',
    remove_old_config=True
)
runner.register(tmux_install_task)

# Zsh
zsh_install_task = create_installer(
    name='zsh',
    template_dir=os.path.join(dir_path, 'zsh'),
    description='Zsh terminal + oh-my-zsh + zdharma',
    default_config_location='~/.zshrc',
    remove_old_config=True
)
runner.register(zsh_install_task)

# Kubectl
kubectl_install_task = create_installer(
    name='kubectl',
    template_dir=os.path.join(dir_path, 'kubectl'),
    description='Kubernetes CLI tool',
    ask_config_location=False
)
runner.register(kubectl_install_task)

# Helm
helm_install_task = create_installer(
    name='helm',
    template_dir=os.path.join(dir_path, 'helm'),
    description='Package manager for kubernetes',
    ask_config_location=False
)
runner.register(helm_install_task)

# Terraform
terraform_install_task = create_installer(
    name='terraform',
    template_dir=os.path.join(dir_path, 'terraform'),
    description='Open source IAC by Hashicorp',
    config_locations=['~/.bashrc', '~/.zshrc']
)
runner.register(terraform_install_task)

# Docker
docker_install_task = create_installer(
    name='docker',
    template_dir=os.path.join(dir_path, 'docker'),
    description='Most popular containerization platform',
    ask_config_location=False
)
runner.register(docker_install_task)



from typing import List
from core.schema import BaseDateTimeSchema, BaseCountSchema


class PascalEntityNameData(BaseDateTimeSchema):
    snake_column_name: str


class PascalEntityName(PascalEntityNameData):
    id: str

    class Config:
        orm_mode = True


class PascalEntityNameResult(BaseCountSchema):
    data: List[PascalEntityName]



from config import (
    app_enable_auth_module
)
from component.log import logger
from component.db_connection import engine
from helper.migration import migrate
from module.auth.component import Base
from module.auth.register_permission import register_permission


async def migrate_auth():
    if not app_enable_auth_module:
        logger.info('🥪 Skip DB migration for "auth"')
        return
    logger.info('🥪 Perform DB migration for "auth"')
    await migrate(engine=engine, Base=Base)
    logger.info('🥪 Register permissions')
    await register_permission()



from typing import List
from core.schema import BaseDateTimeSchema, BaseCountSchema


class ActivityData(BaseDateTimeSchema):
    action: str
    entity: str
    data: str


class Activity(ActivityData):
    id: str

    class Config:
        orm_mode = True


class ActivityResult(BaseCountSchema):
    data: List[Activity]



from typing import Any, List, Optional
from .._common.input import project_dir_input, app_name_input
from ....task.decorator import python_task
from ....task.task import Task
from ....helper import util
from ....helper.codemod.add_import_module import add_import_module
from ....helper.codemod.add_upstream_to_task import add_upstream_to_task
from ....helper.codemod.add_assert_resource import add_assert_resource
from ....helper.file.copy_tree import copy_tree
from ....helper.file.text import read_text_file_async, write_text_file_async
import os

current_dir = os.path.dirname(__file__)


def create_add_project_automation_task(
    upstreams: Optional[List[Task]] = None
) -> Task:
    @python_task(
        name='add-project-automation',
        inputs=[project_dir_input],
        upstreams=[] if upstreams is None else upstreams
    )
    async def _task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir', '.')
        task.print_out('Create project automation modules if not exist')
        await _add_project_automation(project_dir)
    return _task


async def _add_project_automation(project_dir: str):
    if os.path.exists(os.path.join(project_dir, '_automate', '_project')):
        return
    await copy_tree(
        src=os.path.join(current_dir, 'template'),
        dst=project_dir
    )
    project_task_module_path = '_automate._project'
    zrb_init_path = os.path.join(project_dir, 'zrb_init.py')
    code = await read_text_file_async(zrb_init_path)
    import_alias = project_task_module_path.split('.')[-1]
    code = add_import_module(
        code=code,
        module_path=project_task_module_path,
        alias=import_alias
    )
    code = add_assert_resource(code, import_alias)
    await write_text_file_async(zrb_init_path, code)


def create_register_app_task(
    task_name: str,
    project_automation_file_name: str,
    project_automation_task_name: str,
    app_automation_file_name: str,
    app_automation_task_var_name_tpl: str,
    upstreams: Optional[List[Task]] = None,
):
    @python_task(
        name=task_name,
        inputs=[project_dir_input, app_name_input],
        upstreams=[] if upstreams is None else upstreams
    )
    async def _task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir', '.')
        app_name = kwargs.get('app_name')
        snake_app_name = util.to_snake_case(app_name)
        abs_app_automation_file_name = os.path.join(
            project_dir, '_automate', snake_app_name, app_automation_file_name
        )
        app_automation_task_var_name = app_automation_task_var_name_tpl.format(
            snake_app_name=snake_app_name
        )
        project_automation_dir = os.path.join(
            project_dir, '_automate', '_project'
        )
        project_automation_path = os.path.join(
            project_automation_dir, f'{project_automation_file_name}'
        )
        upstream_task_rel_file_path = os.path.relpath(
            abs_app_automation_file_name, project_automation_path
        )
        # normalize `..` parts
        upstream_module_parts = [
            part if part != '..' else ''
            for part in upstream_task_rel_file_path.split(os.path.sep)
        ]
        # remove .py extenstion
        last_part = upstream_module_parts[-1]
        upstream_module_parts[-1] = os.path.splitext(last_part)[0]
        # turn into module path
        upstream_module_path = '.'.join(upstream_module_parts)
        task.print_out(
            f'Add {app_automation_task_var_name} to project automation'
        )
        code = await read_text_file_async(project_automation_path)
        code = add_import_module(
            code=code,
            module_path=upstream_module_path,
            resource=app_automation_task_var_name
        )
        code = add_upstream_to_task(
            code, project_automation_task_name, app_automation_task_var_name
        )
        await write_text_file_async(project_automation_path, code)
    return _task



from zrb.builtin._group import project_group
from zrb import Task, runner

start_project = Task(
    name='start',
    group=project_group,
    upstreams=[],
    description='Start project',
    run=lambda *args, **kwargs: kwargs.get('_task').print_out('👌')
)
runner.register(start_project)



from typing import List
from zrb import CmdTask, Env, EnvFile, runner
from zrb.builtin._group import project_group
from .image import push_snake_app_name_image
from ._common import (
    CURRENT_DIR, DEPLOYMENT_DIR, APP_TEMPLATE_ENV_FILE_NAME,
    DEPLOYMENT_TEMPLATE_ENV_FILE_NAME, MODULES, image_input,
    pulumi_stack_input, mode_input, image_env, pulumi_backend_url_env,
    pulumi_config_passphrase_env
)
import os
import jsons

deployment_app_env_file = EnvFile(
    env_file=APP_TEMPLATE_ENV_FILE_NAME,
    prefix='DEPLOYMENT_APP_ENV_PREFIX'
)
deployment_env_file = EnvFile(
    env_file=DEPLOYMENT_TEMPLATE_ENV_FILE_NAME,
    prefix='DEPLOYMENT_CONFIG_ENV_PREFIX'
)

deployment_modules_env = Env(
    name='MODULES',
    os_name='DEPLOYMENT_CONFIG_ENV_PREFIX_MODULES',
    default=jsons.dumps(MODULES)
)

deployment_envs: List[Env] = [
    pulumi_backend_url_env,
    pulumi_config_passphrase_env,
    image_env,
    deployment_modules_env
]

deployment_mode_env = Env(
    name='MODE',
    os_name='DEPLOYMENT_CONFIG_ENV_PREFIX_MODE',
    default='{{input.snake_app_name_mode}}'
)


###############################################################################
# Task Definitions
###############################################################################

deploy_snake_app_name = CmdTask(
    icon='🚧',
    name='deploy-kebab-app-name',
    description='Deploy human readable app name',
    group=project_group,
    inputs=[
        image_input,
        pulumi_stack_input,
        mode_input,
    ],
    upstreams=[push_snake_app_name_image],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_env_file,
        deployment_app_env_file,
    ],
    envs=deployment_envs + [
        deployment_mode_env,
    ],
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'pulumi-up.sh'),
)
runner.register(deploy_snake_app_name)

destroy_snake_app_name = CmdTask(
    icon='💨',
    name='destroy-kebab-app-name',
    description='Remove human readable app name deployment',
    group=project_group,
    inputs=[
        pulumi_stack_input,
    ],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_env_file,
        deployment_app_env_file,
    ],
    envs=deployment_envs,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'pulumi-destroy.sh'),
)
runner.register(destroy_snake_app_name)



import libcst as cst


def add_function_call(code: str, function_name: str, parameters: list) -> str:
    """
    Adds a function call statement to the end of a module with the specified
    function name and parameters.

    Args:
        code (str): The code of the module to modify.
        function_name (str): The name of the function to call.
        parameters (list): A list of parameters to pass to the function call.

    Returns:
        str: The modified code with the new function call statement added.
    """

    module = cst.parse_module(code)

    # Create the new function call statement
    new_function_call = cst.Expr(
        value=cst.Call(
            func=cst.Name(value=function_name),
            args=[
                cst.Arg(value=cst.parse_expression(param))
                for param in parameters
            ]
        )
    )

    module.body.append(cst.SimpleStatementLine([new_function_call]))

    # Generate the modified code
    modified_code = module.code
    return modified_code



from zrb import DockerComposeTask, HTTPChecker, Env, runner
from zrb.builtin._group import project_group
import os

###############################################################################
# Constants
###############################################################################

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, '..'))
RESOURCE_DIR = os.path.join(PROJECT_DIR, 'src', 'kebab-task-name')

###############################################################################
# Task Definitions
###############################################################################

snake_task_name = DockerComposeTask(
    name='kebab-task-name',
    description='human readable task name',
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd='composeCommand',
    compose_env_prefix='ENV_PREFIX',
    envs=[
        Env(
            name='MESSAGE',
            os_name='ENV_PREFIX_MESSAGE',
            default='Salve Mane'
        ),
        Env(
            name='CONTAINER_PORT',
            os_name='ENV_PREFIX_CONTAINER_PORT',
            default='3000'
        ),
        Env(
            name='HOST_PORT',
            os_name='ENV_PREFIX_HOST_PORT',
            default='httpPort'
        ),
    ],
    checkers=[
        HTTPChecker(
            name='check-kebab-task-name',
            port='{{env.HOST_PORT}}'
        )
    ]
)
runner.register(snake_task_name)



from typing import Any
from .._common.input import (
    project_dir_input, app_name_input, module_name_input, entity_name_input,
    column_name_input, column_type_input
)
from .._common.helper import validate_project_dir
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....runner import runner
from ....helper import util
from ....helper.codemod.add_property_to_class import add_property_to_class
from ....helper.codemod.add_key_value_to_dict import add_key_value_to_dict
from ....helper.file.text import read_text_file_async, write_text_file_async

import asyncio
import os
import re

current_dir = os.path.dirname(__file__)


@python_task(
    name='validate',
    inputs=[
        project_dir_input, app_name_input, module_name_input, entity_name_input
    ],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory does not exist: {automation_dir}'
        )
    app_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if not os.path.exists(app_dir):
        raise Exception(f'App directory does not exist: {app_dir}')
    module_path = os.path.join(app_dir, 'src', 'module', snake_module_name)
    if not os.path.exists(module_path):
        raise Exception(f'Module directory does not exist: {module_path}')
    entity_path = os.path.join(module_path, 'entity', snake_entity_name)
    if not os.path.exists(entity_path):
        raise Exception(f'Entity directory does not exist: {entity_path}')


@python_task(
    name='fastapp-field',
    group=project_add_group,
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        column_name_input,
        column_type_input,
    ],
    upstreams=[
        validate,
    ],
    runner=runner
)
async def add_fastapp_field(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    column_name = kwargs.get('column_name')
    column_type = kwargs.get('column_name')
    kebab_app_name = util.to_kebab_case(app_name)
    kebab_module_name = util.to_kebab_case(module_name)
    snake_module_name = util.to_snake_case(module_name)
    kebab_entity_name = util.to_kebab_case(entity_name)
    snake_entity_name = util.to_snake_case(entity_name)
    pascal_entity_name = util.to_pascal_case(entity_name)
    snake_column_name = util.to_snake_case(column_name)
    kebab_column_name = util.to_kebab_case(column_name)
    capitalized_human_readable_column_name = util.to_capitalized_human_readable( # noqa
        column_name
    )
    await asyncio.gather(
        asyncio.create_task(add_column_to_test(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, snake_column_name, column_type
        )),
        asyncio.create_task(add_column_to_schema(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, pascal_entity_name, snake_column_name,
            column_type
        )),
        asyncio.create_task(add_column_to_repo(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, pascal_entity_name, snake_column_name,
            column_type
        )),
        asyncio.create_task(add_column_to_list_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_detail_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_delete_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_update_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_insert_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
    )
    task.print_out('Success')


async def add_column_to_insert_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'new', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add default value to insert page')
    default_value_regex = r"(.*)(// DON'T DELETE: set field default value here)"
    default_value_subst = '\\n'.join([
        f"\\1row.{snake_column_name} = '';",
        '\\1\\2'
    ])
    html_content = re.sub(
        default_value_regex, default_value_subst, html_content, 0, re.MULTILINE
    )
    task.print_out('Add field to insert page')
    field_regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    field_subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <input type="text" class="input w-full" id="{kebab_column_name}" placeholder="{capitalized_human_readable_column_name}" bind:value=' + '{row.' + snake_column_name + '} />', # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        field_regex, field_subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


async def add_column_to_update_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'update', '[id]', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add field to update page')
    regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <input type="text" class="input w-full" id="{kebab_column_name}" placeholder="{capitalized_human_readable_column_name}" bind:value=' + '{row.' + snake_column_name + '} />', # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        regex, subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


async def add_column_to_delete_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'delete', '[id]', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add field to delete page')
    regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <span id="{kebab_column_name}">' + '{row.' + snake_column_name + '}</span>', # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        regex, subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


async def add_column_to_detail_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    kebab_column_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, 'detail', '[id]', '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    task.print_out('Add field to detail page')
    regex = r"(.*)(<!-- DON'T DELETE: insert new field here-->)"
    subst = '\\n'.join([
        '\\1<div class="mb-4">',
        f'\\1    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{capitalized_human_readable_column_name}</label>', # noqa
        f'\\1    <span id="{kebab_column_name}">' + '{row.' + snake_column_name + '}</span>',  # noqa
        '\\1</div>',
        '\\1\\2'
    ])
    html_content = re.sub(
        regex, subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


async def add_column_to_list_page(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    kebab_module_name: str,
    kebab_entity_name: str,
    snake_column_name: str,
    capitalized_human_readable_column_name: str
):
    list_page_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'frontend', 'src', 'routes',
        kebab_module_name, kebab_entity_name, '+page.svelte'
    )
    task.print_out(f'Read HTML from: {list_page_file_path}')
    html_content = await read_text_file_async(list_page_file_path)
    # process header
    task.print_out('Add column header to table')
    header_regex = r"(.*)(<!-- DON'T DELETE: insert new column header here-->)"
    header_subst = '\\n'.join([
        f'\\1<th>{capitalized_human_readable_column_name}</th>',
        '\\1\\2'
    ])
    html_content = re.sub(
        header_regex, header_subst, html_content, 0, re.MULTILINE
    )
    # process column
    task.print_out('Add column to table')
    column_regex = r"(.*)(<!-- DON'T DELETE: insert new column here-->)"
    column_subst = '\\n'.join([
        '\\1<td>{row.' + snake_column_name + '}</td>',
        '\\1\\2'
    ])
    html_content = re.sub(
        column_regex, column_subst, html_content, 0, re.MULTILINE
    )
    task.print_out(f'Write modified HTML to: {list_page_file_path}')
    await write_text_file_async(list_page_file_path, html_content)


async def add_column_to_test(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    test_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'test', snake_module_name,
        f'test_{snake_entity_name}.py'
    )
    task.print_out(f'Read code from: {test_file_path}')
    code = await read_text_file_async(test_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the test'
    )
    dict_names = [
        'inserted_success_data', 'to_be_updated_success_data',
        'updated_success_data', 'to_be_deleted_success_data'
    ]
    for dict_name in dict_names:
        code = add_key_value_to_dict(
            code=code,
            dict_name=dict_name,
            key=f"'{snake_column_name}'",
            value="''"
        )
    task.print_out(f'Write modified code to: {test_file_path}')
    await write_text_file_async(test_file_path, code)


async def add_column_to_schema(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    pascal_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    schema_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'schema', f'{snake_entity_name}.py'
    )
    task.print_out(f'Read code from: {schema_file_path}')
    code = await read_text_file_async(schema_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the schema'
    )
    code = add_property_to_class(
        code=code,
        class_name=f'{pascal_entity_name}Data',
        property_name=snake_column_name,
        property_type='str'
    )
    task.print_out(f'Write modified code to: {schema_file_path}')
    await write_text_file_async(schema_file_path, code)


async def add_column_to_repo(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    snake_entity_name: str,
    pascal_entity_name: str,
    snake_column_name: str,
    column_type: str
):
    repo_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'module', snake_module_name,
        'entity', snake_entity_name, 'repo.py'
    )
    task.print_out(f'Read code from: {repo_file_path}')
    code = await read_text_file_async(repo_file_path)
    task.print_out(
        f'Add column "{snake_column_name}" to the repo'
    )
    code = add_property_to_class(
        code=code,
        class_name=f'DBEntity{pascal_entity_name}',
        property_name=snake_column_name,
        property_type='Column',
        property_value='Column(String)'
    )
    task.print_out(f'Write modified code to: {repo_file_path}')
    await write_text_file_async(repo_file_path, code)



from typing import Any, Callable, Iterable, List
from ...builtin._group import dev_tool_install_group
from ..cmd_task import CmdTask
from ..task import Task
from ...task_input.base_input import BaseInput
from ...task_input.str_input import StrInput
from ...helper.file.text import read_text_file_async, write_text_file_async

import os

dir_path = os.path.dirname(__file__)


def _get_append_config(
    source_path: str, destination_path: str
) -> Callable[..., Any]:
    async def _append_config(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        if not source_path or not os.path.exists(source_path):
            task.print_out('Nothing to configure')
            return
        destination = task.render_str(
            os.path.expandvars(
                os.path.expanduser(destination_path)
            )
        )
        task.print_out(f'Add configuration to {destination}')
        additional_content = await read_text_file_async(source_path)
        content = await read_text_file_async(destination)
        new_content = content + '\n' + additional_content
        await write_text_file_async(destination, new_content)
        task.print_out('👌')
    return _append_config


def _get_cmd_path(template_dir: str, step: str) -> str:
    file_path = os.path.join(dir_path, template_dir, f'{step}.sh')
    if os.path.exists(file_path):
        return file_path
    return os.path.join(dir_path, '_default', f'{step}.sh')


def create_installer(
    name: str,
    template_dir: str,
    description: str,
    inputs: List[BaseInput] = [],
    ask_config_location: bool = True,
    config_locations: Iterable[str] = [],
    default_config_location: str = '',
    remove_old_config: bool = False,
    executable: str = 'bash'
) -> Task:
    # define new inputs
    inputs = list(inputs)
    if ask_config_location:
        config_hint = ''
        if len(config_locations) > 0:
            config_hint = ' (e.g.,' + ', '.join(config_locations) + ')'
        inputs = inputs + [
            StrInput(
                name='config-file',
                shortcut='c',
                description=f'Config file{config_hint}',
                prompt=f'Config file{config_hint}',
                default=default_config_location
            )
        ]
    # define task chain
    check_task = CmdTask(
        name=f'check-{name}',
        inputs=inputs,
        preexec_fn=None,
        cmd_path=_get_cmd_path(template_dir, 'check'),
        executable=executable
    )
    download_task = CmdTask(
        name=f'download-{name}',
        inputs=inputs,
        upstreams=[check_task],
        preexec_fn=None,
        cmd_path=_get_cmd_path(template_dir, 'download'),
        executable=executable
    )
    backup_config_task = CmdTask(
        name=f'backup-config-{name}',
        inputs=inputs,
        upstreams=[check_task],
        preexec_fn=None,
        cmd_path=_get_cmd_path(template_dir, 'backup-config'),
        executable=executable
    )
    setup_task = CmdTask(
        name=f'setup-{name}',
        inputs=inputs,
        upstreams=[download_task],
        preexec_fn=None,
        cmd_path=_get_cmd_path(template_dir, 'setup'),
        executable=executable
    )
    remove_config_task = CmdTask(
        name='remove-{name}-config',
        inputs=inputs,
        upstreams=[backup_config_task],
        preexec_fn=None,
        cmd_path=_get_cmd_path(template_dir, 'remove-config'),
        executable=executable
    )
    configure_task = Task(
        name=f'configure-{name}',
        inputs=inputs,
        upstreams=[
            setup_task,
            remove_config_task if remove_old_config else backup_config_task
        ],
        run=_get_append_config(
            os.path.join(template_dir, 'config.sh'),
            '{{ input.config_file }}'
        )
    )
    finalize_task = CmdTask(
        name=name,
        description=description,
        group=dev_tool_install_group,
        inputs=inputs,
        upstreams=[configure_task],
        preexec_fn=None,
        cmd_path=_get_cmd_path(template_dir, 'finalize'),
        executable=executable
    )
    return finalize_task



import os

new_task_lock = os.path.sep.join([
    '{{ os.path.join(input.project_dir, "_automate") }}',
    '{{ util.to_snake_case(input.task_name) }}.py'
])

new_task_app_lock = os.path.sep.join([
    '{{ os.path.join(input.project_dir, "_automate") }}',
    '{{ util.to_snake_case(input.app_name) }}.py'
])



from typing import Iterable
from typeguard import typechecked
from ..task.base_task import BaseTask


@typechecked
class BaseAction():

    def __init__(self):
        self._tasks: Iterable[BaseTask] = []

    def register(self, task: BaseTask):
        self._tasks.append(task)



from typing import Any, Callable, List, Optional

TComparator = Callable[[Any, Any], bool]


def ensure_uniqueness(
    data: List[Any], comparator: Optional[TComparator] = None
) -> List[Any]:
    unique_data = []
    for datum in data:
        if datum in unique_data:
            continue
        if comparator is not None:
            is_exist = False
            for unique_datum in unique_data:
                if comparator(unique_datum, datum):
                    is_exist = True
                    break
            if is_exist:
                continue
        unique_data.append(datum)
    return unique_data



from typing import (
    Any, Callable, Iterable, List, Mapping, Optional, TypeVar, Union
)
from typeguard import typechecked
from ..advertisement import advertisements
from ..task_env.env import Env
from ..task_env.env_file import EnvFile
from ..task_input.base_input import BaseInput
from ..task_input._constant import RESERVED_INPUT_NAMES
from ..helper.accessories.color import (
    get_random_color, is_valid_color, colored
)
from ..helper.accessories.icon import get_random_icon
from ..helper.advertisement import get_advertisement
from ..helper.list.ensure_uniqueness import ensure_uniqueness
from ..helper.list.reverse import reverse
from ..helper.log import logger
from ..helper.render_data import DEFAULT_RENDER_DATA
from ..helper.string.double_quote import double_quote
from ..helper.string.conversion import (
    to_cmd_name, to_variable_name, to_boolean
)
from ..helper.map.conversion import to_str as str_map_to_str
from ..helper.string.jinja import is_probably_jinja
from ..config.config import show_advertisement

import asyncio
import copy
import datetime
import inspect
import jinja2
import os
import sys
import time


MAX_NAME_LENGTH = 20
MULTILINE_INDENT = ' ' * 8

TTask = TypeVar('TTask', bound='BaseTask')
TGroup = TypeVar('TGroup', bound='Group')


@typechecked
class Group():
    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        parent: Optional[TGroup] = None
    ):
        self.name = name
        self.description = description
        self.parent = parent
        if parent is not None:
            parent.children.append(self)
        self.children: List[TGroup] = []
        self.tasks: List[TTask] = []

    def get_cmd_name(self) -> str:
        return to_cmd_name(self.name)

    def get_complete_name(self) -> str:
        cmd_name = self.get_cmd_name()
        if self.parent is None:
            return cmd_name
        parent_cmd_name = self.parent.get_complete_name()
        return f'{parent_cmd_name} {cmd_name}'

    def get_id(self) -> str:
        group_id = self.get_cmd_name()
        if self.parent is None:
            return group_id
        parent_group_id = self.parent.get_id()
        return f'{parent_group_id} {group_id}'


class AnyExtensionFileSystemLoader(jinja2.FileSystemLoader):
    def get_source(self, environment, template):
        for search_dir in self.searchpath:
            file_path = os.path.join(search_dir, template)
            if os.path.exists(file_path):
                with open(file_path, 'r') as file:
                    contents = file.read()
                return contents, file_path, lambda: False
        raise jinja2.TemplateNotFound(template)


@typechecked
class TimeTracker():

    def __init__(self):
        self._start_time: float = 0
        self._end_time: float = 0

    def start_timer(self):
        self._start_time = time.time()

    def end_timer(self):
        self._end_time = time.time()

    def get_elapsed_time(self) -> float:
        return self._end_time - self._start_time


@typechecked
class AttemptTracker():

    def __init__(self, retry: int = 2):
        self.retry = retry
        self._attempt: int = 1

    def get_max_attempt(self) -> int:
        return self.retry + 1

    def get_attempt(self) -> int:
        return self._attempt

    def increase_attempt(self):
        self._attempt += 1

    def should_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt <= max_attempt

    def is_last_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt >= max_attempt


@typechecked
class FinishTracker():

    def __init__(self):
        self._execution_queue: Optional[asyncio.Queue] = None
        self._counter = 0

    async def mark_start(self):
        if self._execution_queue is None:
            self._execution_queue = asyncio.Queue()
        self._counter += 1

    async def mark_as_done(self):
        # Tracker might be started several times
        # However, when the execution is marked as done, it applied globally
        # Thus, we need to send event as much as the counter.
        for i in range(self._counter):
            await self._execution_queue.put(True)

    async def is_done(self) -> bool:
        while self._execution_queue is None:
            await asyncio.sleep(0.05)
        return await self._execution_queue.get()


@typechecked
class PidModel():

    def __init__(self):
        self.zrb_task_pid: int = os.getpid()

    def set_task_pid(self, pid: int):
        self.zrb_task_pid = pid

    def get_task_pid(self) -> int:
        return self.zrb_task_pid


@typechecked
class TaskModel(
    PidModel, FinishTracker, AttemptTracker, TimeTracker
):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
    ):
        # init properties
        self.name = name
        self.group = group
        self.envs = envs
        self.env_files = env_files
        self.icon = icon
        self.color = color
        # init parent classes
        PidModel.__init__(self)
        FinishTracker.__init__(self)
        TimeTracker.__init__(self)
        non_negative_retry = self.ensure_non_negative(
            retry, 'Find negative retry'
        )
        AttemptTracker.__init__(self, retry=non_negative_retry)
        # init private properties
        self._input_map: Mapping[str, Any] = {}
        self._env_map: Mapping[str, str] = {}
        self._complete_name: Optional[str] = None
        self._filled_complete_name: Optional[str] = None
        self._rendered_str: Mapping[str, str] = {}
        self._is_keyval_set = False  # Flag
        self._has_cli_interface = False
        self._render_data: Optional[Mapping[str, Any]] = None
        self._all_inputs: Optional[List[BaseInput]] = None

    def get_icon(self) -> str:
        if self.icon is None or self.icon == '':
            self.icon = get_random_icon()
        return self.icon

    def get_color(self) -> str:
        if self.color is None or not is_valid_color(self.color):
            self.color = get_random_color()
        return self.color

    def get_cmd_name(self) -> str:
        return to_cmd_name(self.name)

    def ensure_non_negative(self, value: float, error_label: str) -> float:
        if value < 0:
            self.log_warn(f'{error_label}: {value}')
            return 0
        return value

    def show_done_info(self):
        complete_name = self._get_complete_name()
        elapsed_time = self.get_elapsed_time()
        message = '\n'.join([
            f'{complete_name} completed in {elapsed_time} seconds',
        ])
        self.print_out_dark(message)
        self.play_bell()

    def log_debug(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.debug(colored_message)

    def log_warn(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.warning(colored_message)

    def log_info(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.info(colored_message)

    def log_error(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.error(colored_message, exc_info=True)

    def log_critical(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.critical(colored_message, exc_info=True)

    def print_out(self, msg: Any):
        prefix = self._get_colored_print_prefix()
        print(f'🤖 ➜  {prefix} • {msg}'.rstrip(), file=sys.stderr)

    def print_err(self, msg: Any):
        prefix = self._get_colored_print_prefix()
        print(f'🤖 ⚠  {prefix} • {msg}'.rstrip(), file=sys.stderr)

    def print_out_dark(self, msg: Any):
        self.print_out(colored(msg, attrs=['dark']))

    def colored(self, text: str) -> str:
        return colored(text, color=self.get_color())

    def play_bell(self):
        print('\a', end='', file=sys.stderr)

    def _get_colored_print_prefix(self) -> str:
        return self.colored(self._get_print_prefix())

    def _get_print_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=True)
        icon = self.get_icon()
        truncated_name = self._get_filled_complete_name()
        return f'{common_prefix} • {icon} {truncated_name}'

    def _get_log_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=False)
        icon = self.get_icon()
        filled_name = self._get_filled_complete_name()
        return f'{common_prefix} • {icon} {filled_name}'

    def _get_common_prefix(self, show_time: bool) -> str:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        pid = self.get_task_pid()
        if show_time:
            now = datetime.datetime.now().isoformat()
            return f'{now} ⚙ {pid} ➤ {attempt} of {max_attempt}'
        return f'⚙ {pid} ➤ {attempt} of {max_attempt}'

    def _get_filled_complete_name(self) -> str:
        if self._filled_complete_name is not None:
            return self._filled_complete_name
        complete_name = self._get_complete_name()
        self._filled_complete_name = complete_name.rjust(MAX_NAME_LENGTH, ' ')
        return self._filled_complete_name

    def get_input_map(self) -> Mapping[str, Any]:
        return self._input_map

    def get_env_map(self) -> Mapping[str, str]:
        return self._env_map

    def _inject_env_map(
        self, env_map: Mapping[str, str], override: bool = False
    ):
        for key, val in env_map.items():
            if override or key not in self._env_map:
                self._env_map[key] = val

    def render_any(self, val: Any) -> Any:
        if isinstance(val, str):
            return self.render_str(val)
        return val

    def render_float(self, val: Union[str, float]) -> float:
        if isinstance(val, str):
            return float(self.render_str(val))
        return val

    def render_int(self, val: Union[str, int]) -> int:
        if isinstance(val, str):
            return int(self.render_str(val))
        return val

    def render_bool(self, val: Union[str, bool]) -> bool:
        if isinstance(val, str):
            return to_boolean(self.render_str(val))
        return val

    def render_str(self, val: str) -> str:
        if val in self._rendered_str:
            return self._rendered_str[val]
        if not is_probably_jinja(val):
            return val
        template = jinja2.Template(val)
        data = self._get_render_data()
        self.log_debug(f'Render string template: {val}, with data: {data}')
        try:
            rendered_text = template.render(data)
            self.log_debug(f'Get rendered result: {rendered_text}')
        except Exception:
            raise Exception(f'Fail to render "{val}" with data: {data}')
        self._rendered_str[val] = rendered_text
        return rendered_text

    def render_file(self, location: str) -> str:
        location_dir = os.path.dirname(location)
        env = jinja2.Environment(
            loader=AnyExtensionFileSystemLoader([location_dir])
        )
        template = env.get_template(location)
        data = self._get_render_data()
        data['TEMPLATE_DIR'] = location_dir
        self.log_debug(f'Render template file: {template}, with data: {data}')
        rendered_text = template.render(data)
        self.log_debug(f'Get rendered result: {rendered_text}')
        return rendered_text

    def _get_render_data(self) -> Mapping[str, Any]:
        if self._render_data is not None:
            return self._render_data
        render_data = dict(DEFAULT_RENDER_DATA)
        render_data.update({
            'env': self._env_map,
            'input': self._input_map,
        })
        self._render_data = render_data
        return render_data

    def _get_multiline_repr(self, text: str) -> str:
        lines_repr: Iterable[str] = []
        lines = text.split('\n')
        if len(lines) == 1:
            return lines[0]
        for index, line in enumerate(lines):
            line_number_repr = str(index + 1).rjust(4, '0')
            lines_repr.append(f'{MULTILINE_INDENT}{line_number_repr} | {line}')
        return '\n' + '\n'.join(lines_repr)

    async def _set_local_keyval(
        self,
        kwargs: Mapping[str, Any],
        env_prefix: str = ''
    ):
        if self._is_keyval_set:
            return True
        self._is_keyval_set = True
        # Add self.inputs to input_map
        self.log_info('Set input map')
        for task_input in self.get_all_inputs():
            map_key = self._get_normalized_input_key(task_input.name)
            self._input_map[map_key] = self.render_any(
                kwargs.get(map_key, task_input.default)
            )
        self.log_debug(
            f'Input map:\n{str_map_to_str(self._input_map, item_prefix="  ")}'
        )
        # Construct envs based on self.env_files and self.envs,
        # - self.env_files should have lower priority then self.envs
        # - First self.envs/self.env_files should be overriden by the next
        self.log_info('Merging task envs, task env files, and native envs')
        envs: List[Env] = self._deduplicate_env(self.envs)
        for env_file in reverse(self.env_files):
            envs += env_file.get_envs()
        for env_name in os.environ:
            envs.append(Env(name=env_name, os_name=env_name, renderable=False))
        envs = ensure_uniqueness(envs, self._compare_env_name)
        # Add envs to env_map
        self.log_info('Set env map')
        for task_env in envs:
            env_name = task_env.name
            if env_name in self._env_map:
                continue
            env_value = task_env.get(env_prefix)
            if task_env.renderable:
                env_value = self.render_any(env_value)
            self._env_map[env_name] = env_value
        self.log_debug(
            f'Env map:\n{str_map_to_str(self._env_map, item_prefix="  ")}'
        )

    def _deduplicate_env(self, envs: Iterable[Env]) -> List[Env]:
        # If two environment with the same name exists, the second one should
        # override the first one. But the order of the declaration should not
        # be changed
        return reverse(ensure_uniqueness(
            reverse(envs), self._compare_env_name
        ))

    def _compare_env_name(self, first_env: Env, second_env: Env) -> bool:
        return first_env.name == second_env.name

    def get_all_inputs(self) -> Iterable[BaseInput]:
        # Override this method!!!
        return self._all_inputs

    def _get_normalized_input_key(self, key: str) -> str:
        if key in RESERVED_INPUT_NAMES:
            return key
        return to_variable_name(key)

    def _get_complete_name(self) -> str:
        if self._complete_name is not None:
            return self._complete_name
        executable_prefix = ''
        if self._has_cli_interface:
            executable_prefix += self._get_executable_name() + ' '
        cmd_name = self.get_cmd_name()
        if self.group is None:
            self._complete_name = f'{executable_prefix}{cmd_name}'
            return self._complete_name
        group_cmd_name = self.group.get_complete_name()
        self._complete_name = f'{executable_prefix}{group_cmd_name} {cmd_name}'
        return self._complete_name

    def _get_executable_name(self) -> str:
        if len(sys.argv) > 0 and sys.argv[0] != '':
            return os.path.basename(sys.argv[0])
        return 'zrb'

    def set_has_cli_interface(self):
        self._has_cli_interface = True


@typechecked
class BaseTask(TaskModel):
    '''
    Base class for all tasks.
    Every task definition should be extended from this class.
    '''

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[BaseInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[TTask] = [],
        checkers: Iterable[TTask] = [],
        checking_interval: float = 0,
        retry: int = 2,
        retry_interval: float = 1,
        run: Optional[Callable[..., Any]] = None,
        skip_execution: Union[bool, str] = False
    ):
        TaskModel.__init__(
            self,
            name=name,
            group=group,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            retry=retry
        )
        retry_interval = self.ensure_non_negative(
            retry_interval, 'Find negative retry_interval'
        )
        checking_interval = self.ensure_non_negative(
            checking_interval, 'Find negative checking_interval'
        )
        if checking_interval == 0:
            checking_interval = 0.05 if len(checkers) == 0 else 0.1
        if group is not None:
            group.tasks.append(self)
        self.inputs = inputs
        self.description = description
        self.retry_interval = retry_interval
        self.upstreams = upstreams
        self.checkers = checkers
        self.checking_interval = checking_interval
        self.skip_execution = skip_execution
        self._is_check_triggered: bool = False
        self._is_checked: bool = False
        self._is_execution_triggered: bool = False
        self._is_execution_started: bool = False
        self._run_function: Optional[Callable[..., Any]] = run
        self._args: List[Any] = []
        self._kwargs: Mapping[str, Any] = {}

    async def run(self, *args: Any, **kwargs: Any) -> Any:
        '''
        Do task execution
        Please override this method.
        '''
        if self._run_function is not None:
            if inspect.iscoroutinefunction(self._run_function):
                return await self._run_function(*args, **kwargs)
            return self._run_function(*args, **kwargs)
        return True

    async def check(self) -> bool:
        '''
        Return true when task is considered completed.
        By default, this will wait the task execution to be completed.
        You can override this method.
        '''
        return await self.is_done()

    def get_all_inputs(self) -> Iterable[BaseInput]:
        ''''
        Getting all inputs of this task and all its upstream, non-duplicated.
        '''
        if self._all_inputs is not None:
            return self._all_inputs
        inputs: Iterable[BaseInput] = []
        for upstream in self.upstreams:
            upstream_inputs = upstream.get_all_inputs()
            inputs += upstream_inputs
        inputs += self.inputs
        self._all_inputs = ensure_uniqueness(
            inputs, lambda x, y: x.name == y.name
        )
        return inputs

    def get_description(self) -> str:
        if self.description != '':
            return self.description
        return self.name

    def create_main_loop(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., Any]:
        self.log_info('Create main loop')

        def main_loop(*args: Any, **kwargs: Any) -> Any:
            self.log_info('Copy task')
            self_cp = copy.deepcopy(self)
            return asyncio.run(self_cp._run_and_check_all(
                env_prefix=env_prefix,
                raise_error=raise_error,
                args=args,
                kwargs=kwargs
            ))
        return main_loop

    async def _run_and_check_all(
        self,
        env_prefix: str,
        raise_error: bool,
        args: Iterable[Any],
        kwargs: Mapping[str, Any]
    ):
        self.start_timer()
        self.log_info('Set keyval')
        await self._set_keyval(kwargs=kwargs, env_prefix=env_prefix)
        self.log_info('Get new kwargs')
        new_kwargs = self.get_input_map()
        # make sure args and kwargs['_args'] are the same
        self.log_info('Get new args')
        new_args = copy.deepcopy(args)
        if len(args) == 0 and '_args' in kwargs:
            new_args = kwargs['_args']
        new_kwargs['_args'] = new_args
        # inject self as kwargs['_task']
        new_kwargs['_task'] = self
        self._args = new_args
        self._kwargs = new_kwargs
        # run the task
        try:
            coroutines = [
                asyncio.create_task(self._loop_check(show_info=True)),
                asyncio.create_task(self._run_all(*new_args, **new_kwargs))
            ]
            results = await asyncio.gather(*coroutines)
            result = results[-1]
            self._print_result(result)
            return result
        except Exception as e:
            self.log_critical(f'{e}')
            if raise_error:
                raise
        finally:
            self.play_bell()

    def _print_result(self, result: Any):
        '''
        Print result to stdout so that it can be processed further.
        e.g.: echo $(zrb show principle solid) > solid-principle.txt

        You need to override this method
        if you want to show the result differently.
        '''
        if result is None:
            return
        print(result)

    async def _loop_check(self, show_info: bool = False) -> bool:
        self.log_info('Start checking')
        while not await self._cached_check():
            self.log_debug('Task is not ready')
            await asyncio.sleep(self.checking_interval)
        self.end_timer()
        self.log_info('Task is ready')
        if show_info:
            if show_advertisement:
                selected_advertisement = get_advertisement(advertisements)
                selected_advertisement.show()
            self.show_done_info()
            self.show_run_command()
        return True

    def show_run_command(self):
        params: List[str] = [double_quote(arg) for arg in self._args]
        for task_input in self.get_all_inputs():
            key = task_input.name
            kwarg_key = self._get_normalized_input_key(key)
            quoted_value = double_quote(str(self._kwargs[kwarg_key]))
            params.append(f'--{key} {quoted_value}')
        run_cmd = self._get_complete_name()
        run_cmd_with_param = run_cmd
        if len(params) > 0:
            param_str = ' '.join(params)
            run_cmd_with_param += ' ' + param_str
        colored_run_cmd = colored(f'{run_cmd_with_param}', color='yellow')
        colored_label = colored('To run again: ', attrs=['dark'])
        print(colored(f'{colored_label}{colored_run_cmd}'), file=sys.stderr)

    async def _cached_check(self) -> bool:
        if self._is_check_triggered:
            self.log_debug('Waiting checking flag to be set')
            while not self._is_checked:
                await asyncio.sleep(0.1)
            return True
        self._is_check_triggered = True
        self.log_debug('Start checking')
        check_result = await self._check()
        if check_result:
            self._is_checked = True
            self.log_debug('Set checking flag to True')
        return check_result

    async def _check(self) -> bool:
        '''
        Check current task readiness.
        - If self.checkers is defined,
          this will return True once every self.checkers is completed
        - Otherwise, this will return check method's return value.
        '''
        if len(self.checkers) == 0:
            return await self.check()
        self.log_debug('Waiting execution to be started')
        while not self._is_execution_started:
            # Don't start checking before the execution itself has been started
            await asyncio.sleep(0.1)
        check_coroutines: Iterable[asyncio.Task] = []
        for checker_task in self.checkers:
            check_coroutines.append(
                asyncio.create_task(checker_task._run_all())
            )
        await asyncio.gather(*check_coroutines)
        return True

    async def _run_all(self, *args: Any, **kwargs: Any) -> Any:
        self.log_info('Start running')
        await self.mark_start()
        coroutines: Iterable[asyncio.Task] = []
        # Add upstream tasks to processes
        for upstream_task in self.upstreams:
            coroutines.append(asyncio.create_task(
                upstream_task._run_all(**kwargs)
            ))
        # Add current task to processes
        coroutines.append(self._cached_run(*args, **kwargs))
        # Wait everything to complete
        results = await asyncio.gather(*coroutines)
        return results[-1]

    async def _cached_run(self, *args: Any, **kwargs: Any) -> Any:
        if self._is_execution_triggered:
            self.log_debug('Skip execution because execution flag is True')
            return
        self.log_debug('Set execution flag to True')
        self._is_execution_triggered = True
        self.log_debug('Start running')
        # get upstream checker
        upstream_check_processes: Iterable[asyncio.Task] = []
        for upstream_task in self.upstreams:
            upstream_check_processes.append(asyncio.create_task(
                upstream_task._loop_check()
            ))
        # wait all upstream checkers to complete
        await asyncio.gather(*upstream_check_processes)
        # mark execution as started, so that checkers can start checking
        self._is_execution_started = True
        if self.render_bool(self.skip_execution):
            self.log_info(
                f'Skip execution because config: {self.skip_execution}'
            )
            await self.mark_as_done()
            return None
        # start running task
        result: Any
        local_kwargs = dict(kwargs)
        local_kwargs['_task'] = self
        while self.should_attempt():
            try:
                self.log_debug(
                    f'Invoke run with args: {args} and kwargs: {local_kwargs}'
                )
                result = await self.run(*args, **local_kwargs)
                break
            except Exception:
                if self.is_last_attempt():
                    raise
                attempt = self.get_attempt()
                self.log_error(f'Encounter error on attempt {attempt}')
                self.increase_attempt()
                await asyncio.sleep(self.retry_interval)
        await self.mark_as_done()
        return result

    async def _set_keyval(self, kwargs: Mapping[str, Any], env_prefix: str):
        # if input is not in input_map, add default values
        for task_input in self.get_all_inputs():
            key = self._get_normalized_input_key(task_input.name)
            if key in kwargs:
                continue
            kwargs[key] = task_input.default
        # set current task local keyval
        await self._set_local_keyval(kwargs=kwargs, env_prefix=env_prefix)
        # get new_kwargs for upstream and checkers
        new_kwargs = copy.deepcopy(kwargs)
        new_kwargs.update(self._input_map)
        upstream_coroutines = []
        # set uplstreams keyval
        for upstream_task in self.upstreams:
            upstream_coroutines.append(asyncio.create_task(
                upstream_task._set_keyval(
                    kwargs=new_kwargs, env_prefix=env_prefix
                )
            ))
        # set checker keyval
        local_env_map = self.get_env_map()
        checker_coroutines = []
        for checker_task in self.checkers:
            checker_task.inputs += self.inputs
            checker_task._inject_env_map(local_env_map, override=True)
            checker_coroutines.append(asyncio.create_task(
                checker_task._set_keyval(
                    kwargs=new_kwargs, env_prefix=env_prefix
                )
            ))
        # wait for checker and upstreams
        coroutines = checker_coroutines + upstream_coroutines
        await asyncio.gather(*coroutines)



from typing import Any, Optional, Union
from typeguard import typechecked

from .base_input import BaseInput


@typechecked
class FloatInput(BaseInput):

    def __init__(
        self,
        name: str,
        shortcut: Optional[str] = None,
        default: Optional[Any] = None,
        description: Optional[str] = None,
        show_default: Union[bool, str, None] = None,
        prompt: Union[bool, str] = True,
        confirmation_prompt: Union[bool, str] = False,
        prompt_required: bool = True,
        hide_input: bool = False,
        is_flag: Optional[bool] = None,
        flag_value: Optional[Any] = None,
        multiple: bool = False,
        count: bool = False,
        allow_from_autoenv: bool = True,
        hidden: bool = False,
        show_choices: bool = True,
        show_envvar: bool = False,
        nargs: int = 1
    ):
        BaseInput.__init__(
            self,
            name=name,
            shortcut=shortcut,
            default=default,
            description=description,
            show_default=show_default,
            prompt=prompt,
            confirmation_prompt=confirmation_prompt,
            prompt_required=prompt_required,
            hide_input=hide_input,
            is_flag=is_flag,
            flag_value=flag_value,
            multiple=multiple,
            count=count,
            allow_from_autoenv=allow_from_autoenv,
            type=float,
            hidden=hidden,
            show_choices=show_choices,
            show_envvar=show_envvar,
            nargs=nargs,
        )



from typing import List, Optional
from module.auth.schema.permission import PermissionData
from module.auth.component.model.permission_model import permission_model

DEFAULT_ACTION_NAMES = ['get', 'get_by_id', 'insert', 'update', 'delete']
ACTION_DESCRIPTION = {
    'get': 'list all',
    'get_by_id': 'get single',
    'insert': 'insert new',
    'update': 'update',
    'delete': 'delete',
}


async def ensure_entity_permission(
    module_name: str,
    entity_name: str,
    action_names: Optional[List[str]] = None
):
    if action_names is None:
        action_names = DEFAULT_ACTION_NAMES
    for action_name in action_names:
        caption = ACTION_DESCRIPTION[action_name]
        await permission_model.ensure_permission(PermissionData(
            name=f'{module_name}:{entity_name}:{action_name}',
            description=f'Allow {caption} {entity_name} on {module_name}'
        ))


async def register_permission():
    await ensure_entity_permission(
        module_name='auth', entity_name='group'
    )
    await ensure_entity_permission(
        module_name='auth', entity_name='permission'
    )
    await ensure_entity_permission(
        module_name='auth', entity_name='user'
    )
    await ensure_entity_permission(
        module_name='log',
        entity_name='activity',
        action_names=['get', 'get_by_id']
    )



from fastapi import FastAPI, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from prometheus_fastapi_instrumentator import Instrumentator
from config import (
    app_name, app_enable_frontend, cors_allow_credentials, cors_allow_headers,
    cors_allow_methods, cors_allow_origin_regex, cors_allow_origins,
    cors_expose_headers, cors_max_age
)
from component.app_lifespan import app_state
from component.app_lifespan import app_lifespan
from component.frontend_index import frontend_index_response

app = FastAPI(title=app_name, lifespan=app_lifespan)
Instrumentator().instrument(app).expose(app)

if app_enable_frontend:
    @app.middleware("http")
    async def catch_all(request, call_next):
        response = await call_next(request)
        if response.status_code != 404:
            return response
        api_error = str(response.headers.get('api-error', '')).lower() == 'yes'
        if api_error:
            return response
        return frontend_index_response


app.add_middleware(
    CORSMiddleware,
    allow_origins=cors_allow_origins,
    allow_origin_regex=cors_allow_origin_regex,
    allow_methods=cors_allow_methods,
    allow_headers=cors_allow_headers,
    allow_credentials=cors_allow_credentials,
    expose_headers=cors_expose_headers,
    max_age=cors_max_age,
)


@app.head('/liveness')
@app.get('/liveness')
def get_application_liveness_status():
    '''
    Get application liveness status.
    Will return HTTP response status 200 if application is alive,
    or return 503 otherwise.
    Orchestrator like Kubernetes will restart
    any application with non-healthy liveness status.
    '''
    if app_state.get_liveness():
        return JSONResponse(
            content={'app': app_name, 'alive': True},
            status_code=status.HTTP_200_OK
        )
    return JSONResponse(
        content={'message': 'Service is not alive'},
        status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
        headers={'api-error': 'yes'}
    )


@app.head('/readiness')
@app.get('/readiness')
def get_application_readiness_status():
    '''
    Get application readiness status.
    Will return HTTP response status 200 if application is ready,
    or return 503 otherwise.
    Orchestrator like Kubernetes will only send user request
    to any application with healthy readiness status.
    '''
    if app_state.get_readiness():
        return JSONResponse(
            content={'app': app_name, 'ready': True},
            status_code=status.HTTP_200_OK
        )
    return JSONResponse(
        content={'message': 'Service is not ready'},
        status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
        headers={'api-error': 'yes'}
    )



from typing import Optional, List, Mapping
from core.messagebus.messagebus import Publisher
from module.log.core.historical_repo_model import HistoricalRepoModel
from module.auth.schema.user import (
    User, UserData, UserResult, UserLogin
)
from module.auth.entity.user.repo import UserRepo
from module.auth.schema.token import (
    AccessTokenData, RefreshTokenData, TokenResponse
)
from module.auth.core import AccessTokenUtil, RefreshTokenUtil
from module.auth.entity.permission.model import PermissionModel


class UserModel(
    HistoricalRepoModel[User, UserData, UserResult]
):
    schema_result_cls = UserResult
    log_entity_name = 'user'

    def __init__(
        self,
        repo: UserRepo,
        publisher: Publisher,
        permission_model: PermissionModel,
        access_token_util: AccessTokenUtil,
        access_token_expire_seconds: int,
        refresh_token_util: RefreshTokenUtil,
        refresh_token_expire_seconds: int,
        guest_user: User,
        admin_user: Optional[User] = None,
        admin_user_password: str = ''
    ):
        super().__init__(repo, publisher)
        self.permission_model = permission_model
        self.access_token_util = access_token_util
        self.access_token_expire_seconds = access_token_expire_seconds
        self.refresh_token_util = refresh_token_util
        self.refresh_token_expire_seconds = refresh_token_expire_seconds
        self.guest_user = guest_user
        self.admin_user = admin_user
        self.admin_user_pasword = admin_user_password

    async def insert(self, data: UserData) -> User:
        if (
            self.admin_user is not None and
            data.username == self.admin_user.username
        ):
            raise ValueError(
                f'Invalid username: {data.username} is used by admin'
            )
        if data.username == self.guest_user.username:
            raise ValueError(
                f'Invalid username: {data.username} is used by guest'
            )
        return await super().insert(data)

    async def update(self, id: str, data: UserData) -> User:
        if self.admin_user is not None:
            if self.is_admin(id):
                raise ValueError(
                    'Forbidden: editting admin user is not permitted'
                )
            if data.username == self.admin_user.username:
                raise ValueError(
                    f'Invalid username: {data.username} is used by admin'
                )
        if self.is_guest(id):
            raise ValueError(
                'Forbidden: editting guest user is not permitted'
            )
        if data.username == self.guest_user.username:
            raise ValueError(
                f'Invalid username: {data.username} is used by guest'
            )
        return await super().update(id, data)

    async def delete(self, id: str) -> User:
        if self.admin_user is not None and self.is_admin(id):
            raise ValueError(
                'Forbidden: deleting admin user is not permitted'
            )
        if self.is_guest(id):
            raise ValueError(
                'Forbidden: deleting guest user is not permitted'
            )
        return await super().delete(id)

    async def get_by_id(self, id: str) -> User:
        if self.admin_user is not None and self.is_admin(id):
            return self.admin_user
        if id == self.guest_user.id:
            return self.guest_user
        return await super().get_by_id(id)

    async def is_authorized(
        self, id: str, *permission_names: str
    ) -> Mapping[str, bool]:
        user = await self.get_by_id(id)
        if self.is_admin(user.id):
            return {
                permission_name: True for permission_name in permission_names
            }
        user_permission_names = self._get_permission_names(user)
        return {
            permission_name: permission_name in user_permission_names
            for permission_name in permission_names
        }

    def _get_permission_names(self, user: User) -> List[str]:
        permission_names = [
            permission.name for permission in user.permissions
        ]
        for group in user.groups:
            additional_permission_names = [
                permission.name
                for permission in group.permissions
                if permission.name not in permission_names
            ]
            permission_names += additional_permission_names
        return permission_names

    def is_guest(self, id: str) -> bool:
        return id == self.guest_user.id

    def is_admin(self, id: str) -> bool:
        if self.admin_user is None:
            return False
        return id == self.admin_user.id

    async def create_auth_token(self, user_login: UserLogin) -> TokenResponse:
        user = await self._get_user_by_user_login(user_login)
        return TokenResponse(
            access_token=self._get_access_token(user),
            refresh_token=self._get_refresh_token(user),
            token_type='bearer'
        )

    async def refresh_auth_token(
        self, refresh_token: str, access_token: str
    ) -> TokenResponse:
        access_token_data = self.access_token_util.decode(
            access_token, parse_expired_token=True
        )
        refresh_token_data = self.refresh_token_util.decode(refresh_token)
        if access_token_data.user_id != refresh_token_data.user_id:
            raise ValueError('Unmatch refresh and access token')
        user = await self.get_by_id(refresh_token_data.user_id)
        return TokenResponse(
            access_token=self._get_access_token(user),
            refresh_token=self._get_refresh_token(user),
            token_type='bearer'
        )

    def _get_access_token(self, user: User) -> str:
        access_token_data = AccessTokenData(
            user_id=user.id,
            username=user.username,
            expire_seconds=self.access_token_expire_seconds
        )
        return self.access_token_util.encode(access_token_data)

    def _get_refresh_token(self, user: User) -> str:
        refresh_token_data = RefreshTokenData(
            user_id=user.id,
            expire_seconds=self.refresh_token_expire_seconds
        )
        return self.refresh_token_util.encode(refresh_token_data)

    async def _get_user_by_user_login(self, user_login: UserLogin) -> User:
        if user_login.identity == '':
            raise ValueError('Invalid identity: Identity is empty')
        if (
            self.admin_user is not None
            and user_login.password == self.admin_user_pasword
            and (
                user_login.identity == self.admin_user.username or
                user_login.identity == self.admin_user.email or
                user_login.identity == self.admin_user.phone
            )
        ):
            return self.admin_user
        return await self.repo.get_by_user_login(user_login)



from typing import List
from ....task.task import Task
from ....task.cmd_task import CmdTask
from .._common.input import (
    project_dir_input, app_name_input, module_name_input, entity_name_input,
)

import os

current_dir = os.path.dirname(__file__)
codemod_dir = os.path.join(current_dir, 'nodejs', 'codemod')

nav_config_file_path = '{{input.project_dir}}/src/{{util.to_kebab_case(input.app_name)}}/src/frontend/src/lib/config/navData.ts' # noqa
nav_var_name = 'navData'
nav_title = '{{util.to_pascal_case(input.entity_name)}}'
nav_url = '/{{util.to_kebab_case(input.module_name)}}/{{util.to_kebab_case(input.entity_name)}}' # noqa
nav_permission = '{{util.to_snake_case(input.module_name)}}:{{util.to_snake_case(input.entity_name)}}:get' # noqa


def create_add_navigation_task(upstreams: List[Task]) -> Task:
    return CmdTask(
        name='add-navigation',
        inputs=[
            project_dir_input,
            app_name_input,
            module_name_input,
            entity_name_input,
        ],
        upstreams=upstreams,
        retry=0,
        cmd=f'node {codemod_dir}/dist/addNav.js "{nav_config_file_path}" "{nav_var_name}" "{nav_title}" "{nav_url}" "{nav_permission}"' # noqa
    )



from module.log.core.historical_repo_model import HistoricalRepoModel
from module.auth.schema.group import (
    Group, GroupData, GroupResult
)


class GroupModel(
    HistoricalRepoModel[Group, GroupData, GroupResult]
):
    schema_result_cls = GroupResult
    log_entity_name = 'group'



from typing import Any, Mapping
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller, Server
from core.repo import SearchFilter
from module.log.component.model.activity_model import (
    activity_model
)
from module.log.schema.activity import ActivityData
from module.auth.schema.token import AccessTokenData


def register_rpc(
    logger: Logger,
    rpc_server: Server,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "log.activity"')

    @rpc_server.register('log_get_activity')
    async def get(
        keyword: str,
        criterion: Mapping[str, Any],
        limit: int,
        offset: int,
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        result = await activity_model.get(
            search_filter=SearchFilter(
                keyword=keyword, criterion=criterion
            ),
            limit=limit,
            offset=offset
        )
        return result.dict()

    @rpc_server.register('log_get_activity_by_id')
    async def get_by_id(
        id: str,
        user_token_data: Mapping[str, Any] = {}
    ) -> Mapping[str, Any]:
        row = await activity_model.get_by_id(id)
        return row.dict()

    @rpc_server.register('log_insert_activity')
    async def insert(
        data: Mapping[str, Any],
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data['created_by'] = user_token_data.user_id
        data['updated_by'] = user_token_data.user_id
        row = await activity_model.insert(
            data=ActivityData(**data)
        )
        return row.dict()

    @rpc_server.register('log_update_activity')
    async def update(
        id: str,
        data: Mapping[str, Any],
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data['updated_by'] = user_token_data.user_id
        row = await activity_model.update(
            id=id, data=ActivityData(**data)
        )
        return row.dict()

    @rpc_server.register('log_delete_activity')
    async def delete(
        id: str,
        user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        row = await activity_model.delete(id=id)
        return row.dict()



from module.auth.core import Authorizer, RPCAuthorizer
from component.rpc import rpc_caller

authorizer: Authorizer = RPCAuthorizer(
    rpc_caller=rpc_caller,
    is_admin_rpc_name='auth_user_is_admin',
    is_guest_rpc_name='auth_user_is_guest',
    is_user_authorized_rpc_name='auth_is_user_authorized'
)



from fastapi import FastAPI, Depends
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller
from core.error import HTTPAPIException
from module.auth.core import Authorizer
from module.auth.schema.group import (
    Group, GroupData, GroupResult
)
from module.auth.schema.token import AccessTokenData
from module.auth.component import access_token_scheme


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register API for "auth.group"')

    @app.get(
        '/api/v1/auth/groups', response_model=GroupResult
    )
    async def get_groups(
        keyword: str = '', limit: int = 100, offset: int = 0,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:permission:get'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_get_group',
                keyword=keyword,
                criterion={},
                limit=limit,
                offset=offset,
                user_token_data=user_token_data.dict()
            )
            return GroupResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get(
        '/api/v1/auth/groups/{id}', response_model=Group
    )
    async def get_group_by_id(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:group:get_by_id'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_get_group_by_id',
                id=id, user_token_data=user_token_data.dict()
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post(
        '/api/v1/auth/groups', response_model=Group
    )
    async def insert_group(
        data: GroupData,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:group:insert'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_insert_group',
                data=data.dict(), user_token_data=user_token_data.dict()
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put(
        '/api/v1/auth/groups/{id}', response_model=Group
    )
    async def update_group(
        id: str, data: GroupData,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:group:update'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_update_group',
                id=id, data=data.dict(), user_token_data=user_token_data.dict()
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete(
        '/api/v1/auth/groups/{id}', response_model=Group
    )
    async def delete_group(
        id: str,
        user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, 'auth:group:delete'
        ):
            raise HTTPAPIException(403, 'Unauthorized')
        try:
            result_dict = await rpc_caller.call(
                'auth_delete_group',
                id=id, user_token_data=user_token_data.dict()
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



from sqlalchemy import Column, String, Table, ForeignKey
from module.auth.component import Base

user_group = Table(
    'user_groups',
    Base.metadata,
    Column('user_id', String, ForeignKey('users.id')),
    Column('group_id', String, ForeignKey('groups.id')),
)

user_permission = Table(
    'user_permissions',
    Base.metadata,
    Column('user_id', String, ForeignKey('users.id')),
    Column('permission_id', String, ForeignKey('permissions.id')),
)

group_permission = Table(
    'group_permissions',
    Base.metadata,
    Column('group_id', String, ForeignKey('groups.id')),
    Column('permission_id', String, ForeignKey('permissions.id')),
)



from zrb import DockerComposeTask, Env, HTTPChecker, runner
from zrb.builtin._group import project_group
from ._common import (
    RESOURCE_DIR,
    local_input, host_input, https_input, image_input,
    image_env
)
from .image import build_snake_app_name_image

compose_env_prefix = 'CONTAINER_ENV_PREFIX'
compose_envs = [
    image_env,
    Env(
        name='HOST_PORT',
        os_name='CONTAINER_ENV_PREFIX_HOST_PORT',
        default='httpPort'
    ),
]

###############################################################################
# Task Definitions
###############################################################################

remove_snake_app_name_container = DockerComposeTask(
    icon='💨',
    name='remove-kebab-app-name-container',
    description='Rumove human readable app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd='down',
    compose_env_prefix=compose_env_prefix,
    envs=compose_envs,
)
runner.register(remove_snake_app_name_container)

stop_snake_app_name_container = DockerComposeTask(
    icon='⛔',
    name='stop-kebab-app-name-container',
    description='Stop human readable app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd='stop',
    compose_env_prefix=compose_env_prefix,
    envs=compose_envs,
)
runner.register(stop_snake_app_name_container)

start_snake_app_name_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-app-name-container',
    description='Start human readable app name container',
    group=project_group,
    inputs=[
        local_input,
        host_input,
        https_input,
        image_input,
    ],
    skip_execution='{{not input.local_snake_app_name}}',
    upstreams=[
        build_snake_app_name_image,
        remove_snake_app_name_container
    ],
    cwd=RESOURCE_DIR,
    compose_cmd='up',
    compose_env_prefix=compose_env_prefix,
    envs=compose_envs,
    checkers=[
        HTTPChecker(
            name='check-kebab-app-name',
            host='{{input.snake_app_name_host}}',
            port='{{env.HOST_PORT}}',
            is_https='{{input.snake_app_name_https}}'
        )
    ]
)
runner.register(start_snake_app_name_container)



from config import app_logging_level
import logging

# create logger
logger = logging.getLogger('src')
logger.setLevel(app_logging_level)

ch = logging.StreamHandler()
ch.setLevel(app_logging_level)

# create formatter
formatter = logging.Formatter(
    '%(levelname)s:\t%(message)s'
)

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)
